/*
 *	fhist - file history and comparison tools
 *	Copyright (C) 1991, 1992, 1993, 1994 Peter Miller.
 *	All rights reserved.
 *
 *	Derived from a work
 *	Copyright (C) 1990 David I. Bell.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * MANIFEST: functions to prune away old edits from the edit history file.
 */

#include <stdio.h>
#include <ac/string.h>

#include <breaks.h>
#include <compare.h>
#include <error.h>
#include <fhist.h>
#include <fileio.h>
#include <prune.h>
#include <subroutine.h>


/*
 * Prune old edits from the history file to save on disk space.
 * This removes all edits up until the specified edit number.
 *
 * The history (.e) file is binary, because we need to seek in it.
 * The source (.s) file is text, because we don't need to seek in it.
 * The input files are text, by definition.
 * The output files are text, by definition.
 */

void
prunehistory(editname)
	char	*editname;	/* edit name to stop pruning at */
{
	FILE	*fp;		/* input history file */
	FILE	*ofp;		/* output history file */
	POS	*pp;		/* current entry in position table */
	long	lastprunededit;	/* last edit number to be pruned */
	long	firstkeepedit;	/* first edit number to keep */
	long	prunecount;	/* number of edits being pruned */
	long	copypos;	/* starting position to copy from */
	long	newtablepos;	/* position of new position table */
	long	edit;		/* current edit number */
	long	shift;		/* amount to shift data up by */
	char	tempname[1024]; /* temporary name of new edit file */

	strcpy(tempname, sc.basename);
	strcat(tempname, EXT_NEW);
	fp = openhistoryfile(OHF_READ);
	firstkeepedit = findeditnumber(fp, editname);
	lastprunededit = firstkeepedit - 1;
	prunecount = firstkeepedit - sc.firstedit;
	if (prunecount <= 0)
	{
		fclose(fp);
		if (fc.verbosity)
			error("No edits pruned from module \"%s\"", sc.modulename);
		return;
	}
	if (!sc.forceupdateflag)
	{
		printf("The edit history for module \"%s\" currently has edits %ld to %ld.\n",
			sc.modulename, sc.firstedit, sc.lastedit);
		printf("Do you wish to irretrievably discard edit");
		printf((sc.firstedit == lastprunededit) ? " %ld? " : "s %ld to %ld? ",
			sc.firstedit, lastprunededit);
		if (queryuser())
			return;
	}
	pp = readpostable(fp);
	pp++;			/* ignore unused entry */

	/*
	 * This will be the new history file,
	 * so create it binary.
	 */
	ofp = fopen(tempname, "w+b");
	if (!ofp)
		nfatal("create \"%s\"", tempname);
	fprintf(ofp, HEADERFORMAT, T_HEADER, 0L, 0L, 0L, 0L);
	copypos = pp[sc.lastedit - firstkeepedit].p_pos;
	seekf(fp, copypos);
	copyf(fp, ofp, sc.tablepos - copypos);
	if (fclose(fp))
		nfatal("read \"%s\"", sc.historyname);
	newtablepos = ftell(ofp);
	shift = sc.tablepos - newtablepos;
	for (edit = sc.lastedit; edit >= firstkeepedit; edit--, pp++)
		fprintf(ofp, (pp->p_names ? "%c %ld %ld %s\n" : "%c %ld %ld\n"),
			T_POSITION, edit, pp->p_pos - shift, pp->p_names);
	fprintf(ofp, "%c %ld\n", T_EOF, ftell(ofp));
	seekf(ofp, 0L);
	fprintf(ofp, HEADERFORMAT, T_HEADER, firstkeepedit, sc.lastedit,
		0L, newtablepos);
	if (fclose(ofp))
		nfatal("write \"%s\"", tempname);
	breaksoff();
	if (renamefiles(EXT_HISTORY))
		nfatal("Rename failed for edit history file \"%s\"",
			sc.historyname);
	if (fc.verbosity)
		error("Pruned %ld edit%s from module \"%s\"", prunecount,
			(prunecount == 1) ? "" : "s", sc.modulename);
	breakson();
}
