/*
 *	fhist - file history and comparison tools
 *	Copyright (C) 1991, 1992, 1993, 1994 Peter Miller.
 *	All rights reserved.
 *
 *	Derived from a work
 *	Copyright (C) 1990 David I. Bell.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * MANIFEST: functions to append the changes for a new edit to the history file.
 */

#include <stdio.h>
#include <errno.h>
#include <ac/string.h>
#include <ac/time.h>
#include <ac/unistd.h>

#include <breaks.h>
#include <cmalloc.h>
#include <compare.h>
#include <error.h>
#include <fhist.h>
#include <fileio.h>
#include <modlin.h>
#include <subroutine.h>
#include <update.h>


/*
 * Add the lines in the history file for the specified new edit number.
 * This are the beginning and end lines, the file line, and the default
 * and user-supplied remark lines.  These lines are written to the current
 * position in the file.  Errors are remembered by the file structure.
 */

static void addhistory _((FILE *fp, char *remarks, long editnum));

static void
addhistory(fp, remarks, editnumber)
	FILE	*fp;		/* file to write to */
	char	*remarks;	/* remarks to add to file */
	long	editnumber;	/* edit number */
{
	time_t	bintime;	/* binary time of day */
	long	len;		/* length of remark line */

	time(&bintime);
	fprintf(fp, "%c %ld\n", T_BEGINEDIT, editnumber);
	fprintf(fp, "%c %s   %s", T_REMARK, cuserid(NULL), ctime(&bintime));
	while (*remarks)
	{
		len = strlen(remarks);
		fprintf(fp, "%c ", T_REMARK);
		writef(fp, remarks, len);
		remarks += (len + 1);
	}
	fprintf(fp, "%c 0\n", T_FILE);
	fprintf(fp, "%c %ld 0 0\n", T_ENDEDIT, editnumber);
}


/*
 * Describe help about entering remarks.
 */

static void helpremarks _((void));

static void
helpremarks()
{
	printf("\nRemarks are lines of text which describe why an edit was made.\n");
	printf("The lines of remarks are terminated by typing a single blank line.\n");
	printf("Lines beginning with an exclaimation mark are commands, as follows:\n\n");
	printf("!b   Backs up to the previous line so you can retype it.\n");
	printf("!c   Clears all remarks entered so far so you can retype them.\n");
	printf("!r   Redisplays all remarks entered so far.\n");
	printf("!q   Aborts updating of the current module.\n\n");
}


/*
 * Read remark lines from the user to describe the reason why an edit
 * was made.  This is made up of lines separated by nulls, with a double
 * null at the end.  The remarks can be from the terminal or from a file.
 *
 * Remark files are text.
 */

static char *getremarks _((char *modulename, char *remarkname, long editnum));

static char *
getremarks(modulename, remarkname, editnumber)
	char	*modulename;	/* module name this is for */
	char	*remarkname;	/* file name for remarks (NULL if from tty) */
	long	editnumber;	/* edit number */
{
	char	*buf;		/* buffer for input */
	FILE	*fp;		/* remark file */
	char	*cp;		/* current line for retyping */
	long	avail;		/* total characters available */
	long	used;		/* number of characters used */
	long	line;
	int	ch;

	if (remarkname && (*remarkname == '\0'))	/* no remarks at all */
		return "";
	used = 0;
	avail = 400;
	buf = cm_alloc_and_check(avail + 4);

	/*
	 * If remark file name was supplied, read remarks from it.
	 */
	if (remarkname)
	{
		fp = fopen(remarkname, "r");
		if (!fp)
			nfatal("open \"%s\"", remarkname);
		for (;;)
		{
			cp = readlinef(fp, &line, 0);
			if (!cp)
				break;
			if ((used + line) >= avail)
			{
				avail = used + line + 200;
				buf = cm_realloc_and_check(buf, avail + 4);
			}
			strcpy(&buf[used], cp);
			used += (line + 1);
		}
		if (fclose(fp))
			nfatal("close \"%s\"", remarkname);
		buf[used] = '\0';
		return buf;
	}

	/*
	 * No remark file specified, so get remarks from user.
	 */
	printf("\nPlease type remarks for edit %ld of \"%s\" (type \"!h\" for help)\n",
		editnumber, modulename);
	line = 1;
	for (;;)
	{
		if ((used + 200) > avail)
		{
			avail += 200;
			buf = cm_realloc_and_check(buf, avail + 4);
		}
		printf("%ld> ", line);
		fflush(stdout);
		if (fgets(&buf[used], 200, stdin) == NULL)
			break;
		ch = buf[used];
		if (ch == '\n')
			break;
		if (ch != '!')
		{
			/* normal line */
			used += strlen(&buf[used]) + 1;
			line++;
			continue;
		}
		buf[used] = '\0';
		ch = buf[used+1];		/* control line */
		if ((ch >= 'A') && (ch <= 'Z'))
			ch += ('a' - 'A');
		switch (ch)
		{
		case 'b':
			/* back up to previous line */
			if (line <= 1)
				break;
			line--;
			cp = buf;
			for (ch = 1; ch < line; ch++)
				cp += (strlen(cp) + 1);
			used = (cp - buf);
			break;

		case 'c':
			/* restart the remark */
			printf("\nRemarks cleared, please type new remarks:\n");
			used = 0;
			line = 1;
			break;

		case 'h':
			/* help */
			helpremarks();
			break;

		case 'q':
			/* quit the command */
			fatal("no action taken");

		case 'r':
			/* retype remarks */
			printf("\n");
			for (ch = 0, cp = buf; *cp; cp += (strlen(cp) + 1))
				printf("%d> %s", ++ch, cp);
			break;

		default:
			/* unknown command */
			if (ch < ' ') ch = ' ';
			printf("Unknown command \"!%c\" - ignored\n", ch);
		}
	}
	buf[used] = '\0';		/* double null to end string */
	return buf;
}


/*
 * Do the copy of the source file to the destination file, while
 * modifying the lines as necessary.  Returns nonzero on an error.
 */

static int docopy _((FILE *ip, FILE *op));

static int
docopy(ip, op)
	FILE	*ip;		/* input file */
	FILE	*op;		/* output file */
{
	char	*cp;		/* current line */
	long	linelen;	/* current line length */
	long	i;		/* line count */

	for (i = sc.modifylines; i > 0; i--)
	{
		cp = readlinef(ip, &linelen, 0);
		if (cp == NULL)
			return ferror(ip);
		cp = modifyline(cp, &linelen, (INFO *) NULL);
		if (writef(op, cp, linelen))
			return -1;
	}
	return copyf(ip, op, -1L);
}


/*
 * Create a new module.
 * This creates the new history file, and the initial source file.
 * This will start as edit 1, with the given file as the original source.
 *
 * The history (.e) file is binary, because we need to seek in it.
 * The source (.s) file is text, because we don't need to seek in it.
 * The input files are text, by definition.
 * The output files are text, by definition.
 */

void
createhistory(inputname, editname)
	char	*inputname;	/* name of file to input */
	char	*editname;	/* name to give this edit (NULL if none) */
{
	FILE	*fp;		/* history file */
	FILE	*ip;		/* input filename */
	FILE	*up;		/* updated source file */
	long	pospos, editpos;
	char	*remarks;	/* remarks for new file */

	/*
	 * Verify that the module doesn't already exist, and
	 * open up the initial file for reading.
	 */
	fp = fopen(sc.historyname, "rb");
	if (fp)
	{
		fclose(fp);
		fatal("Module \"%s\" already exists as \"%s\"",
			sc.modulename, sc.basename);
	}
	if (errno != ENOENT)
		nfatal("open \"%s\"", sc.historyname);
	up = fopen(sc.sourcename, "r");
	if (up)
	{
		fclose(up);
		fatal("Source for module \"%s\" already exists as \"%s\"",
			sc.modulename, sc.basename);
	}
	if (errno != ENOENT)
		nfatal("open \"%s\"", sc.sourcename);
	ip = fopen(inputname, "r");
	if (!ip)
		nfatal("open \"%s\"", inputname);

	/*
	 * Try to create the history source file, and if successful, then
	 * get the initial remarks for the module.
	 */
	if (fc.verbosity)
		error("Creating new module \"%s\" from initial file \"%s\"",
			sc.modulename, inputname);
	up = fopen(sc.sourcename, "w");
	if (!up)
		nfatal("create \"%s\"", sc.sourcename);
	remarks = getremarks(sc.modulename, sc.remarkname, 1L);

	/*
	 * Copy the original file into the history latest source file,
	 * with the extra line at the front identifying it as edit 1.
	 */
	breaksoff();
	if (fc.verbosity > VERBOSE_DEFAULT)
		error("[Copying \"%s\" to \"%s\"]", inputname, sc.sourcename);
	fprintf(up, "%c 1\n", T_BEGINEDIT);
	if (ferror(up))
		nfatal("write \"%s\"", sc.sourcename);
	if (docopy(ip, up))
		nfatal("Error during copy of \"%s\" to \"%s\"",
			inputname, sc.sourcename);
	if (fclose(up))
		nfatal("write \"%s\"", sc.sourcename);
	if (fclose(ip))
		nfatal("close \"%s\"", inputname);

	/*
	 * Write the initial history file data
	 *	create it binary
	 */
	if (fc.verbosity > VERBOSE_DEFAULT)
		error("[Creating history file \"%s\"]", sc.historyname);
	fp = fopen(sc.historyname, "w+b");
	if (!fp)
		nfatal("create \"%s\"", sc.historyname);
	fprintf(fp, HEADERFORMAT, T_HEADER, 1L, 1L, 0L, 0L);
	editpos = ftell(fp);
	addhistory(fp, remarks, 1L);
	pospos = ftell(fp);
	fprintf(fp, (editname ? "%c 1 %ld %s\n" : "%c 1 %ld\n"),
		T_POSITION, editpos, editname);
	fprintf(fp, "%c %ld\n", T_EOF, ftell(fp));
	seekf(fp, 0L);
	fprintf(fp, HEADERFORMAT, T_HEADER, 1L, 1L, 0L, pospos);
	if (fclose(fp))
		nfatal("write \"%s\"", sc.historyname);
	if (fc.verbosity > VERBOSE_DEFAULT)
	{
		error("[Creation complete]");
		fflush(stdout);
	}
	breakson();
}


/*
 * Write the new source file into a temporary file in the proper directory.
 * This is essentially a copy, except for the initial line indicating the
 * edit number of the new source.  This copy is done to guarantee that there
 * is enough disk space in order to finish the update.  If there is not, then
 * nothing harmful will have been done to the old source or history files.
 * It is assumed that the history file has already been opened.
 *
 * The history (.e) file is binary, because we need to seek in it.
 * The source (.s) file is text, because we don't need to seek in it.
 * The input files are text, by definition.
 * The output files are text, by definition.
 */

static void writesource _((char *inputname));

static void
writesource(inputname)
	char	*inputname;		/* file containing new source */
{
	FILE	*ip;			/* input file */
	FILE	*up;			/* output file */
	char	updatename[1024];	/* full name of new source file */

	strcpy(updatename, sc.basename);
	strcat(updatename, EXT_NEW);
	if (fc.verbosity > VERBOSE_DEFAULT)
		error("[Copying \"%s\" to temporary file \"%s\"]", inputname, updatename);
	ip = fopen(inputname, "r");
	if (!ip)
		nfatal("open \"%s\"", inputname);
	up = fopen(updatename, "w");
	if (!up)
		nfatal("create \"%s\"", updatename);
	fprintf(up, "%c %ld\n", T_BEGINEDIT, sc.lastedit + 1);
	if (ferror(up))
		nfatal("write \"%s\"", updatename);
	if (docopy(ip, up))
		nfatal("copy \"%s\" to \"%s\"", inputname, updatename);
	if (fclose(up))
		nfatal("write \"%s\"", updatename);
	if (fclose(ip))
		nfatal("close \"%s\"", inputname);
}


/*
 * Compare two files, and return the changes into a snake list.
 * This works by calling the fcomp module which is loaded with us.
 * Does not return if an error ocurred.
 */

static void comparefiles _((char *file1, char *file2, long skip2));

static void
comparefiles(file1, file2, skip2)
	char	*file1;		/* first file to compare */
	char	*file2;		/* second file to compare */
	long	skip2;		/* number of lines to skip for second file */
{
	if (fc.verbosity > VERBOSE_DEFAULT)
		error("[Comparing files \"%s\" and \"%s\"]", file1, file2);
	fc.fileBskip = skip2;
	fc.fileAmodify = sc.modifylines;
	fc.fileBmodify = sc.modifylines;
	fc.maxchanges = INFINITY;
	fcomp(file1, file2);
}


/*
 * Swap the snake line numbers for two files to appear as if the comparison
 * was in the other order.  This is used when showing differences and then
 * updating, since showing differences needs the changes to the new version,
 * but updating needs the changes back to the older version.
 */

static void swapfiles _((void));

static void
swapfiles()
{
	SNAKE		*sp;		/* current snake */
	long		temp;		/* temporary for swapping */
	FILEINFO	tempfile;	/* temporary for swapping */

	for (sp = fc.snakelist; sp; sp = sp->next)
	{
		temp = sp->line1;
		sp->line1 = sp->line2;
		sp->line2 = temp;
	}
	tempfile = fc.fileA;
	fc.fileA = fc.fileB;
	fc.fileB = tempfile;
	temp = fc.deletes;
	fc.deletes = fc.inserts;
	fc.inserts = temp;
}


/*
 * Write out the edit script corresponding to the results of the
 * comparison between two files.  Errors cause premature finishing,
 * and are accessible by using ferror afterwards.  This routine knows
 * the data structures defined by the fcomp module.  This routine is
 * also called after the begin line and remark lines have been
 * skipped over.
 */

static void writeedits _((FILE *fp, long editnum));

static void
writeedits(fp, editnumber)
	FILE	*fp;		/* file to write to */
	long	editnumber;	/* edit number to write */
{
	SNAKE	*sp;		/* current snake */
	long	line1;
	long	line2;		/* current line numbers in files */
	long	inserts;
	long	deletes;	/* number of inserts and deletes needed */
	long	totalinserts;
	long	totaldeletes;	/* total lines inserted or deleted */
	long	i;
	char	*cp;		/* line being inserted */

	totalinserts = 0;
	totaldeletes = 0;
	line1 = 0;
	line2 = 0;
	for (sp = fc.snakelist; sp; sp = sp->next)
	{
		deletes = sp->line1 - line1;
		inserts = sp->line2 - line2;
		if (fc.debugflag)
			error("[line1 %ld  line2 %ld  count %ld  insert %ld  delete %ld]",
				sp->line1, sp->line2, sp->count, inserts, deletes);
		if (deletes || inserts)
		{
			fprintf(fp, "%c %ld %ld %ld\n", T_CHANGE, line1 + 1, inserts, deletes);
			if (ferror(fp))
				return;
		}
		for (i = 0; i < inserts; i++)
		{
			fprintf(fp, "%c ", T_TEXT);
			cp = fc.fileB.f_lines[line2 + i]->l_data;
			writef(fp, cp, (long) strlen(cp));
			if (ferror(fp))
				return;
		}
		totalinserts += inserts;
		totaldeletes += deletes;
		line1 = sp->line1 + sp->count;
		line2 = sp->line2 + sp->count;
	}
	fprintf(fp, "%c %ld %ld %ld\n", T_ENDEDIT,
		editnumber, totalinserts, totaldeletes);
}


/*
 * Update the history file by appending the changes given by a file.
 */

void
updatehistory(inputname, editname)
	char	*inputname;	/* name of file to input */
	char	*editname;	/* name to give this edit (NULL if none) */
{
	FILE	*fp;		/* history file */
	char	*cp;		/* current line of data */
	POS	*pp;		/* current entry in position table */
	POS	*postable;	/* position table */
	char	*remarks;	/* remarks */
	long	newtablepos;	/* position of new position table */
	long	lasteditpos;	/* position of beginning of last edit */
	long	appendpos;	/* position to append new changes to */
	long	tempedit;	/* edit number just scanned from line */
	long	newedit;	/* number of new edit */
	long	edit;		/* current edit number */
	int	saveerror;	/* saved error code on errors */

	/*
	 * Make sure that the input file contains the right version number
	 * before beginning the update, and if so, then get the remarks
	 * for the edit.
	 */
	fp = openhistoryfile(OHF_WRITE);
	if (!sc.forceupdateflag)
	{
		tempedit = findedit(inputname);
		if ((tempedit > 0) && (tempedit != sc.lastedit))
		{
			fatal
			(
	 "Input file \"%s\" contains edit %ld instead of %ld for module \"%s\"",
				inputname,
				tempedit,
				sc.lastedit,
				sc.modulename
			);
		}
	}
	newedit = sc.lastedit + 1;
	if (fc.verbosity)
	{
		error
		(
			"Installing \"%s\" as edit %ld of module \"%s\"",
			inputname,
			newedit,
			sc.modulename
		);
	}
	remarks = getremarks(sc.modulename, sc.remarkname, newedit);

	/*
	 * See if we are doing a straight update, or an update after a
	 * display of the differences.  If the latter, then we already have
	 * done the comparison, but we need to swap the differences.
	 */
	if (fc.snakelist)
		swapfiles();
	else
		comparefiles(inputname, sc.sourcename, 1L);
	postable = readpostable(fp);

	/*
	 * Now go back and verify that the last edit actually exists where it
	 * is supposed to be.  We will overwrite it with the changes.
	 * Skip over all remark lines to position to the place to append
	 * the new changes.  This will be at the location of the file line
	 * for the latest edit.
	 */
	lasteditpos = postable[1].p_pos;
	cp = getline(fp, lasteditpos, T_BEGINEDIT);
	if (cp == NULL)
		fatal("Cannot read beginning line of latest edit");
	cp = getnumber(cp, &tempedit);
	if ((cp == NULL) || (tempedit != sc.lastedit))
		fatal("Bad begin edit line for latest edit");

	for (;;)
	{
		appendpos = ftell(fp);
		cp = readlinef(fp, (long *) NULL, 0);
		if (cp == NULL)
			fatal("Error reading remark lines for latest edit");
		if (*cp == T_FILE)	/* the file line, all done */
			break;
		if (*cp != T_REMARK)	/* error if not remark line */
			fatal("Unexpected line type in latest edit");
	}

	/*
	 * Now write the new version of the source file under a temporary name.
	 * This is to prevent problems due to quota errors.  So if this fails
	 * we have done nothing bad.
	 */
	breaksoff();
	writesource(inputname);

	/*
	 * Now position back to the file line in the latest edit, and
	 * overwrite it with the changes for the new edit.  Errors will
	 * be checked for later and handled there, so we can recover from
	 * disk full or quota exceeded problems.
	 */
	if (fc.verbosity > VERBOSE_DEFAULT)
		error("[Appending edits to history file]");
	if (seekf(fp, appendpos))
		fatal("Cannot seek to location to append new edit");
	writeedits(fp, sc.lastedit);

	/*
	 * Now append the latest edit template to the file, and follow
	 * that with the saved position table, and then append a new
	 * end of file line.  a new end of file line.  Save the
	 * position of the beginning of the table for updating the
	 * header.  Errors are checked for later.
	 */
	pp = postable;
	pp->p_pos = ftell(fp);
	pp->p_names = editname;
	addhistory(fp, remarks, newedit);
	newtablepos = ftell(fp);
	for (edit = newedit; edit >= sc.firstedit; edit--, pp++)
	{
		fprintf
		(
			fp,
			(pp->p_names ? "%c %ld %ld %s\n" : "%c %ld %ld\n"),
			T_POSITION,
			edit,
			pp->p_pos,
			pp->p_names
		);
	}
	fprintf(fp, "%c %ld\n", T_EOF, ftell(fp));

	/*
	 * Force the last bit of the output to the file, then check for
	 * errors.  If an error occurred during any of the above writes
	 * then we will try to recover by rewriting the old position
	 * table back to where it was.  This should succeed since quota
	 * exceeded or disk full errors cannot occur since the position
	 * table was there previously!
	 */
	fflush(fp);
	if (ferror(fp) || renamefiles(EXT_SOURCE))
	{
		saveerror = errno;
		errno = 0;
		if (seekf(fp, appendpos))
		{
			nerror("seekf(\"%s\", %ld)", sc.historyname, appendpos);
			fatal("file \"%s\" is damaged", sc.historyname);
		}
		fprintf(fp, "%c 0\n", T_FILE);
		fprintf(fp, "%c %ld 0 0\n", T_ENDEDIT, sc.lastedit);
		if (ftell(fp) != sc.tablepos)
		{
			error("file \"%s\" position table moved", sc.historyname);
			fatal("file \"%s\" is damaged", sc.historyname);
		}
		pp = &postable[1];
		for (edit = sc.lastedit; edit >= sc.firstedit; edit--, pp++)
			fprintf(fp, (pp->p_names ? "%c %ld %ld %s\n" : "%c %ld %ld\n"),
				T_POSITION, edit, pp->p_pos, pp->p_names);
		fprintf(fp, "%c %ld\n", T_EOF, ftell(fp));
		fflush(fp);
		if (ferror(fp))
		{
			nerror("write \"%s\"", sc.historyname);
			fatal("file \"%s\" is damaged", sc.historyname);
		}
		fclose(fp);
		errno = saveerror;
		nerror("write \"%s\"", sc.historyname);
		fatal("file \"%s\" is not damaged", sc.historyname);
	}

	/*
	 * As the last step for inserting the new edit, rewrite the old header
	 * line to update the new last edit number and the new position of the
	 * beginning of the position table.  This line must be of fixed length.
	 * This write should succeed since the file is not being grown in size.
	 */
	if (seekf(fp, 0L))
	{
		nerror("seekf(\"%s\", 0)", sc.historyname);
		fatal("file \"%s\" is damaged", sc.historyname);
	}
	fprintf(fp, HEADERFORMAT, T_HEADER, sc.firstedit, newedit, 0L,
		newtablepos);
	if (ferror(fp) || fflush(fp))
	{
		nerror("write \"%s\"", sc.historyname);
		fatal("file \"%s\" is damaged", sc.historyname);
	}
	fclose(fp);
	if (fc.verbosity)
		error("Edit %ld of module \"%s\" saved (inserts %ld, deletes %ld, matches %ld)",
			newedit, sc.modulename, fc.deletes,
			fc.inserts, fc.matches);
	fflush(stdout);
	breakson();
}
